/**
 * Manages pagination, sorting and filtering state for a Collection subject to
 * JSON API conventions.
 */
class CollectionQueryParameters {
  /**
   * Initialize the query param set
   * @param {Collection} collection - the Backbone collection to which to attach
   */
  constructor(collection) {
    this.collection = collection
  }

  /**
   * The top level links contain pagination information (whether there are next
   * or previous pages). Stores can call into this method when they receive
   * next top level links objects via a fetch, which can ultimately change
   * the value of `hasNext` and `hasPrev`
   * @param {Object} topLevelLinks - A top level links object as described by
   *   the JSON API spec http://jsonapi.org/format/#fetching-pagination
   */
  handleLinks(topLevelLinks) {
    this._links = topLevelLinks
  }

  /**
   * Configure whether this collection should automatically re-fetch data when
   * it receives new query parameters such as page, a filter, or a new sort.
   * @param {Boolean} newValue - The new value of the fetchExplicitly tag.
   *   `true` means that `fetch` must be called manually when desired. `false`
   *    (the default) means a fetch will occur when relevant query params are
   *    updated.
   */
  set fetchExplicitly(newValue) {
    this._fetchExplicitly = newValue
  }

  /**
   * Return the current value of the fetchExplicitly flag. `false` by default
   * i.e. fetch automatically when parameters change.
   */
  get fetchExplicitly() {
    return this._fetchExplicitly || false
  }

  /**
   * ## Filtering
   * See: http://jsonapi.org/format/#fetching-filtering
   */

  /**
   * Get the current state of the filter parameters.
   */
  get filters() {
    this._filters = this._filters || {}
    return this._filters
  }

  /**
   * Set a filter to a new value. If fetchExplicitly is `false` (the default)
   * then this will cause a refetch.
   * @param {String} filterName - the name of the filter to apply. If filterName
   *   is foo, then the fetch will have a parameter called filter[foo] per the
   *   JSON API spec.
   * @param {String} newValue - the new value of the filter to apply
   */
  setFilter(filterName, newValue) {
    // if filter changes, reset pagination
    if (this.filters[filterName] !== newValue) {
      this.resetPage()
    }
    this.filters[filterName] = newValue
    this.maybeFetch()
  }

  /**
   * Get the current value of one filter
   * @param {String} filterName - the name of the filter of which to retrieve
   *   the value
   */
  getFilter(filterName) {
    return this.filters[filterName]
  }

  /**
   * ## Pagination
   * http://jsonapi.org/format/#fetching-pagination
   */

  /**
   * Set the selected page to a new value and fetch is `fetchExplicitly` is
   * `false`
   * @param {Integer} n - the new page number (first page starts at 1)
   */
  setPage(n) {
    this.page = n
  }

  /**
   * Get the currently selected paginaion size
   */
  getSize() {
    return this.size
  }

  /**
   * Set the selected size to a new value and fetch is `fetchExplicitly` is
   * `false`
   * @param {Integer} n - the new size number
   */
  setSize(n) {
    this.size = n
  }

  /**
   * Get the currently selected pagination limit
   */
  getLimit() {
    return this.offset
  }

  /**
   * Set the selected limit to a new value and fetch is `fetchExplicitly` is
   * `false`
   * @param {Integer} n - the new limit number
   */
  setLimit(n) {
    this.limit = n
  }

  /**
   * Get the currently selected page number (1 by default)
   */
  getPage() {
    return this.page
  }

  /**
   * Return `true` if there is a page after the current one. Note: this is based
   * on the links object returned from the last fetch.
   */
  hasNext() {
    return !!(this._links && this._links.next)
  }

  /**
   * Return `true` if there is a page before the current one. Note: this is
   * based on the links object returned from the last fetch.
   */
  hasPrev() {
    return !!(this._links && this._links.previous)
  }

  /**
   * Set the selected page to a new value and fetch is `fetchExplicitly` is
   * `false`
   * @param {Integer} n - the new page number (first page starts at 1)
   */
  set page(n) {
    this._page = n
    this.maybeFetch()
  }

  /**
   * Set the selected size to a new value, limit to null, resets page
   * and fetch is `fetchExplicitly` is `false`
   * https://github.com/cerebris/jsonapi-resources#paged-paginator
   * @param {Integer} n - the new pagination size
   */

  set size(n) {
    this._size = n
    this._limit = null
    this.resetPage()
    this.maybeFetch()
  }

  /**
   * Set the selected limit to a new value, size to null, resets page
   * and fetch is `fetchExplicitly` is `false`
   * https://github.com/cerebris/jsonapi-resources#offset-paginator
   * @param {Integer} n - the new pagination limit
   */

  set limit(n) {
    this._limit = n
    this._size = null
    this.resetPage()
    this.maybeFetch()
  }

  /**
   * Get the currently selected page number (1 by default)
   */
  get page() {
    return this._page || 1
  }

  get offset(){
    if(this.page > 1){
      return (this.page - 1) * this.limit
    }

    return 0;
  }

  /**
   * Get the currently selected size for pagination
   */

  get size() {
    return this._size
  }

  /**
   * Get the currently selected limit for pagination
   */

  get limit() {
    return this._limit
  }

  /**
   * Get the currently selected paginator type:
   * If size exist it returns paged
   * If limit exist it returns offset
   * If no-one exist it returns nil for backward compatibility
   */

  get paginatorType() {
    if(this.size){
      return "paged"
    }

    if(this.limit){
      return "offset"
    }
  }

  /**
   * Set the page back to page 1. Frequently we do this when changing sorting
   * or filtering options so that the user is not disoriented in the result set.
   */
  resetPage() {
    this._page = 1
  }

  /**
   * ## Sorting
   * http://jsonapi.org/format/#fetching-sorting
   */

  /**
   * Set the sort parameter to a new value. Allows `-` for descending sorts,
   * and comma separated lists of fields for compound sorting.
   * @params {String} newValue - the new sorting option in the spec convention
   */
  set sort(newValue) {
    this._sort = newValue
    this.resetPage()
    this.maybeFetch()
  }

  /**
   * Return the current sorting setting
   */
  get sort() {
    return this._sort
  }

  /**
   * ## Fetching
   */

  /**
   * Build an Object suitable for passing as the `data` option to the `fetch`
   * call. Generally, the fetch is done via GET and this parameters become a
   * set of URL encoded query parameters.
   */
  get fetchData() {
    // Pagination options
    let params = {
      page: this.page
    }

    // The size option if set, otherwise skip
    if(this.paginatorType === "paged") {
      params.page = {size: this.size, number: this.page}
    }

    if(this.paginatorType === "offset") {
      params.page = {limit: this.limit, offset: this.offset}
    }

    // The sort option if set, otherwise skip
    if(this.sort) {
      _.extend(params, {sort: this.sort})
    }

    // Each non-null, non empty string, filter value, as filter[name]
    _.each(_.compactObject(this.filters), (value, filterName) => {
      if (!_.isEmpty(value)) params[`filter[${filterName}]`] = value
    })

    return params
  }

  /**
   * Check `fetchExplicitly`. If it is `false` (the default), do a fetch
   */
  maybeFetch() {
    if(!this.fetchExplicitly) {
      this.collection.fetch()
    }
  }

  /**
   * ## Event Handler factories
   * These are convenience methods for associating a
   * filtering/sorting/pagination operation to a browser event.
   * E.g. making an onClick handler for a button that fetches the next page.
   */

  /**
   * Return an event handler that changes the value of the filter `filterName`
   * to the current value of the target element (form field)
   * @param {String} filterName - the name of the filter to update
   */
  filterChangeHandler(filterName) {
    return (e) => {
      e.preventDefault()
      this.setFilter(filterName, e.target.value)
    }
  }

  /**
   * Return an event handler that changes the sort mode to the current value of
   * the target element (form field)
   */
  sortChangeHandler() {
    return (e) => {
      e.preventDefault()
      this.sort = e.target.value
    }
  }

  /**
   * Return an event handler that advances the pagination to the next page
   */
  nextHandler() {
    return (e) => {
      e.preventDefault()
      this.page = this.page + 1
    }
  }

  /**
   * Return an event handler that advances the pagination to the previous page
   */
  previousHandler() {
    return (e) => {
      e.preventDefault()
      this.page = this.page - 1
    }
  }
}
